#!/usr/bin/env python
# Filename : dongler.py

"""Dongler is written by Richard West,   and is designed as a GUI interface
    for the Huawei E220 modem
    Copyright (C) 2009 Richard West 
                                                                                
    richardbwest@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation:w
, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>

"""
try:
    import Tkinter as TK
    import tkMessageBox
    import pickle
    import os
    import sys

    import humod
except:
    ImportError()

# Version = 1.2.1
# Credits
# The tux logo is a modified version, the original being created by
# Keyser -http://tux.crystalxp.net/en.id.2300-tux-wifi-logo.html 


WW = 400 # width of the master window
WH = 175# height of the master window
APP_BG = "white"
TITLE_COLOUR = "grey20"
ITEM_COLOUR = "grey40"
BUTTON_COLOUR = "blue"
INFO_COLOUR = "orange"
HEADER_BG = "#DCFFBB"
DIS_BG = "#DCFFBB"
BUTTON_BG = "#DCFFBB"
MESSAGE_FG = "slate gray"
MESSAGE_BG = "#DCFFBB"
NAV_COLOUR = "Blue"
PN_COLOUR = "darkgreen"
PADX = 12
PADY = 10
MINIMUM_ROWS = 20


class MainWindow:
    """The main window for dongler.

    """
    def __init__(self, master, dongle, database):
        #building the master frame
        self.database = database
        self.dongle = dongle
        self.master = master
        self.master.protocol("WM_DELETE_WINDOW", self.quit_window)
        self.master.after(2000, self.timed_events)
        # create the tux wireless logo
        self.tux = TK.PhotoImage(file = "images/tuxwireless.ppm")
        self.tuxlabel = TK.Label(self.master, height = 100, width = 100, \
                                 image = self.tux, bg = APP_BG, \
                                 borderwidth = 0)
        self.tuxlabel.grid(column = 0, row = 0,rowspan = 2, \
                           padx = PADX, pady = PADY)
        self.update_signal_icon()
        self.update_mail_icon()
        # create blank for signal strength
        #self.sig1 = TK.PhotoImage(file ="images/0.gif")
        #self.siglabel = TK.Label(self.master, height=50, width=50, \
        #    image=self.sig1, bg = APP_BG, borderwidth = 0)
        #self.siglabel.grid(column=3,  row = 0, padx = PADX, pady = PADY)
        
        #button for displaying the phonebook
        self.pb_button = TK.Button(self.master, text = "Phone Book", \
            fg = BUTTON_COLOUR,command = self.phonebook_window,\
            bg = BUTTON_BG, borderwidth=0)
        self.pb_button.grid(column=0, row = 2, padx = PADX, pady = PADY)

        #button for bringing up the inbox
        self.read_button = TK.Button(self.master, text="Inbox", \
            fg = BUTTON_COLOUR, command = self.inbox_window, \
            bg = BUTTON_BG, borderwidth=0)
        self.read_button.grid(column=1, row = 2, padx = PADX, pady=PADY)

        #button for sending sms
        self.send_button = TK.Button(self.master, text= "Send Sms", \
            fg = BUTTON_COLOUR, command = self.send_window, borderwidth=0, \
            bg = BUTTON_BG)
        self.send_button.grid(column=2, row = 2, padx = PADX, pady = PADY)

        #button to kill the program
        self.quit_button = TK.Button(self.master, text="Quit", fg = "red", \
            command = self.quit_window, borderwidth = 0, bg=BUTTON_BG)
        self.quit_button.grid(column=3, row = 2, padx = PADX, pady = PADY)

        if not self.dongle.online:
            info = tkMessageBox.showinfo("Offline Mode", \
            "Unable to load the modem, so working in offline mode")

    def update_signal_icon(self):
        """adds a graphic for the signal strength to the main window\
        and updates the graphic
        """
        s = self.dongle.get_SS()
        if s == 99:
            self.current_file = "images/err.gif" 
        else:
            self.current_file = "images/%d.gif" % s

        self.current_img = TK.PhotoImage(file = self.current_file)
        self.siglabel = TK.Label(self.master, height=50, width=50, \
            image = self.current_img, bg = APP_BG, borderwidth = 0)
        self.siglabel.grid(column=3, row = 0, padx = PADX,\
                        pady = 1, sticky = "n")

    def update_mail_icon(self):
        """displays an icon depending on whether there are unread/read/new
            messages.
        """
        if self.database.data["new message"]:
            self.new_mail()
            return
        elif self.database.has_unread_messages():
            image = "images/mail.gif"
        else:
            image = "images/nomail.gif"
        self.mail_img = TK.PhotoImage(file = image)
        self.mail_label = TK.Label(self.master, height = 50, width = 60, \
                                   image = self.mail_img, bg = APP_BG,\
                                   borderwidth = 0)
        self.mail_label.grid(column = 3, row = 1,padx = PADX, \
                             pady = 1, sticky = "n")

    def new_mail(self, blink = True): 
        """ when new mail arrives, blink the mail icon"""
        if blink:
            image = "images/mail.gif"
        else:
            image = "images/nomail.gif"
        self.mail_img = TK.PhotoImage(file = image)
        self.mail_label = TK.Label(self.master, height = 50, width = 60, \
                                   image = self.mail_img, bg = APP_BG, \
                                   borderwidth = 0)
        self.mail_label.grid(column = 3, row = 1, padx = PADX, \
                             pady = 1, sticky = "n")
        blink = not(blink)
        if self.database.data["new message"]:
            self.mail_after_id = self.master.after(300, self.new_mail, blink)

    def inbox_window(self):
        """Builds the inbox window

        """
        InboxWindow(self, database = self.database)
        self.database.data["new message"] = False

    def send_window(self, recipients = [], message = ""):
        """Builds the send window

        """
        SendWindow(self, recipients = recipients, message= message)
    
    def phonebook_window(self):
        """Builds the phone book window

        """
        Phonebook_Window(self)
    
    def quit_window(self):
        """Intructs the program to terminate

        """
        self.master.quit()
        self.master.destroy()
        self.database.save_database()

    def timed_events(self):
        """all scheduled repeating activities are put under here.
         (e.g. - checking for new msgs)
        """
        #check for change in signal strength
        self.update_signal_icon()
        self.update_mail_icon()
        self.database.update()
        #schedule the next check
        self.master.after(5000, self.timed_events)


class InboxWindow:
    """The window for viewing the inbox.

    """
    def __init__(self, parent, database, sort = "date"):
        self.database = database
        self.sort = sort
        # build the read message window and add widgets
        self.parent = parent
        self.height = 480
        self.width = 680
        self.window = TK.Toplevel(bg = APP_BG)
        self.window.geometry(centre_me(height = self.height,\
                                       width = self.width))
        self.window.title("Inbox")
        self.window.bind("<Button-1>", self.msg_selected)
        self.build_grid(sortby = self.sort,  page = 1)
        
    def build_grid(self, page, sortby = "date", reverse = True):
        """Buils the grid within the Inbox window

        """
        anchor = "w"
        self.sortby = sortby
        self.reverse = reverse
        self.page = page
        #Add headers to the columns
        headwidth = [20, 35, 15, 10]
        for col in range(len(self.database.data["inboxwindowcolumns"])):
            text = self.database.data["inboxwindowcolumns"][col]
            if text == "number":
                text = "from"
            text = text[:1].upper()+text[1:]
            item = TK.Label(self.window, text = text, width = headwidth[col], \
                      bg = HEADER_BG, anchor = anchor, fg = TITLE_COLOUR)
            item.grid(row = 0, column = col, sticky = "w", padx = 10, pady = 2)
            item.whathit = "HED%s" % text

        #check for new messages
        self.database.update()
        #add in the message contents into the rows
        row = 1
        self.rowdict = {}
        #get the messagelist and slice off the contents needed
        start = (self.page * 20) - 20
        end = start + 20
        self.msglist = self.database.messageIDlist(sortby, reverse, \
        filter = ("REC UNREAD", "REC READ"))
        self.pages = len(self.msglist) / 20
        if len(self.msglist) > (self.pages *20):
            self.pages += 1
        slice = self.msglist[start:end]

        for index in slice:
            for col in range(len(self.database.data["inboxwindowcolumns"])):
                colname = self.database.data["inboxwindowcolumns"][col]         
                text = getattr(self.database, "GETMSG_%s"%colname, \
                    self.database.GET_default)(index, client = "Inbox")
                label = TK.Label(self.window, text = text, bg = APP_BG, \
                          anchor = anchor, fg = ITEM_COLOUR)
                label.grid(row = row, column = col, sticky = "w", padx = 10)
                label.whathit = "MSG%d"% index
                self.rowdict[index] = row
                
            row += 1
        #adds a few more filler lines if the list is only small
        if row <= MINIMUM_ROWS:
            for num in range((MINIMUM_ROWS-row)+1):
                label = TK.Label(self.window, text = " ", bg = APP_BG)
                label.grid(row = row, column = 0)
                label.whathit = "filler"
                row += 1
        #Display the page number
        TK.Label(self.window, text = " Page %d of %d"%(self.page, self.pages), \
        bg = APP_BG, anchor = anchor, fg = PN_COLOUR).grid(row = row, \
            column = 0, pady = 10, padx = 5, columnspan = 2, sticky = "w")
        row += 1
        #Display navigation bars.
        label = TK.Label(self.window, text = "<< Previous", bg = APP_BG, \
            anchor = anchor, fg = NAV_COLOUR)
        label.grid(row = row, column = 1, sticky = "w", padx = 10)
        label.whathit = "NAVNEXT"
        label = TK.Label(self.window, text = "Next >>", bg = APP_BG, \
                     anchor = anchor, fg = NAV_COLOUR)
        label.grid(row = row, column = 2, sticky = "w", padx = 10)
        label.whathit = "NAVPREV"
        row += 1
        #add a tip
        TK.Label(self.window, \
        text = "Tip: Click on a message to view the full message.", \
            bg = APP_BG, anchor = anchor, fg = INFO_COLOUR).grid(row = row, \
        column = 0, pady = 10, padx = 5, columnspan = 2, sticky = "w")

        #add the back button
        self.back_button = TK.Button(self.window, text="Back", fg = "red",  \
            command = self.window.destroy, bg = BUTTON_BG, borderwidth = 0)
        self.back_button.grid(row = row, \
            column = (len(self.database.data["inboxwindowcolumns"])-1), \
            pady = 10, padx=10)

    def refresh(self, sortby, page, sortme = False):
        """Refresh the contents of the inbox view grid.
           If the column header clicked is the same, reverse the list
        """
        if sortme == True:
            if sortby.lower() == self.sortby.lower():
                self.reverse = not(self.reverse)
        for item in self.window.grid_slaves():
            item.grid_forget()
        self.build_grid(sortby = sortby, reverse = self.reverse, page = page)

    def msg_selected(self, event):
        """Event handler for when a message or header is selected
           Dispatch to other methods, dependent on what was clicked
        """
        # set up a dispatcher.
        try :
            method, params = event.widget.whathit[:3], \
                               event.widget.whathit[3:]
        except:
            method, params = "DEF", ""
        getattr(self, "%s_clicked"%method, self.DEF_clicked)(params)


    def MSG_clicked(self, params):
        """if a row is clicked, bring up the message details screen"""
        mid = int(params)
        self.parent.database.data["messages"][mid]["status"] = "REC READ"
        self.parent.master.after(500, self.ViewWindow(mid=mid, parent = self))
        self.refresh(self.sortby, page = self.page)

    def ViewWindow(self, parent, mid):
        """Bring up the message view window"""
        ViewWindow(mid = mid, parent = self)
        
    def NAV_clicked(self, params):
        """Deals with Nagivation button clicked events"""
        if params == "NEXT":
            page = self.page - 1
        elif params == "PREV":
            page = self.page + 1

        if page < 1:
            page = 1# make sure it doesnt go into a minux number
        elif page > self.pages:
            page = self.pages
        self.refresh(sortby = self.sortby, sortme = False, page = page)

    def HED_clicked(self, params):
        """if a header is clicked, refresh the srceen.
           Sort by the header clicked
        """
        if params.lower() == "from":
            params = "number"
        self.refresh(sortby = params.lower(), sortme = True, page = self.page)

    def DEF_clicked(self, params = None):
        """Default for the dispatcher

        """
        pass

    def hide_row(self, mid):
        """Removes a row from the grid, if the message is deleted

        """
        for item in self.window.grid_slaves(row = self.rowdict[mid]):
            item.grid_remove()


class SendWindow:
    """The window that deals with typing an sending new messages

    """
    def __init__(self, parent, recipients = [], message="", width = 530, \
                 height = 250):
        self.parent = parent
        self.recipients = recipients
        self.window = TK.Toplevel(bg = APP_BG)
        self.window.geometry(centre_me(width = width, height = height))
        self.window.title("Send a Message")
        #add the Send  button.
        TK.Button(self.window, text = "Send", fg = BUTTON_COLOUR, \
                  bg = BUTTON_BG, anchor = "w", command = self.send)\
                 .grid(column = 4, row = 2, sticky ="e", padx = 10)
        #add the Discard button.
        TK.Button(self.window, text = "Discard", fg = "red", bg = BUTTON_BG, \
               anchor = "w", command = self.window.destroy).grid(column = 5, \
               row = 2, sticky = "e")
        #add the address book listbox and corresponding scroll bar
        sb = TK.Scrollbar(self.window)
        sb.grid(column = 1,  row = 1, sticky="N"+"s"+"e"+"w", padx = 3, pady=3)
        self.li = TK.Listbox(self.window, yscrollcommand=sb.set, \
                          selectmode="extended")
        self.li.config(selectbackground = BUTTON_BG)

        # get the names from the database and populate the list
        self.namelist = self.parent.database.get_contact_list()
        for name in self.namelist:
            self.li.insert(TK.END, name)
        for number in recipients:
            name = self.parent.database.has_a_name(number)
            if name:
                self.li.selection_set((self.namelist.index(name)))
            else:
                self.li.insert(0, number)
                self.namelist.insert(0, number)
                self.li.selection_set((self.namelist.index(number)))
                self.parent.database.edit_contact(name = number,\
                                                number = number)

        self.li.grid(column = 0, row = 1, sticky = "N"+"s"+"e"+"w", \
                     padx = 3, pady =3)
        sb.config(command = self.li.yview)

        #add the recipient label:
        TK.Label(self.window, text = "Send to:",  fg = TITLE_COLOUR, \
                  bg = APP_BG).grid(column = 0, row = 0)

        #add the recipient box
        self.to_box = TK.Text(self.window)
        self.to_box.config(state = "disabled", foreground = "black", \
                           background = DIS_BG, width = 45, height = 2)
        self.to_box.grid(column = 2,  row = 0, sticky = "w"+"e", columnspan = 4)

        #add the message content box 
        self.message = TK.Text(self.window)
        self.message.insert(TK.END, message)
        self.message.config(height = 11, width = 45)
        self.message.grid(column = 2, row = 1, padx = 3, pady=3, columnspan = 4)
        #add a tool tip
        TK.Label(self.window, \
              text = " Multiple Recipients -  Use Ctrl or Shift select", \
              fg = INFO_COLOUR, bg = APP_BG)\
             .grid(column = 0, row = 2, columnspan=3, sticky="w")
        self.update()

    def update(self):
        """updates the To: entry widget with contact names

        """
        #change the widget state to normal to allow changes
        self.to_box.config(state = "normal")
        self.to_box.delete(1.0, TK.END)
        indexlist = self.li.curselection()#get the selection indexes
        selected = []
        for index in indexlist:
            selected.append(self.namelist[int(index)])
        names = "\"" + "\", \"".join(selected) + "\""
        if len(names) == 2:
            names = ""#tidy up names if the list is empty
        self.to_box.insert(TK.END, names)
        self.to_box.config(state = "disabled")# disable editing again
        self.window.after(250, self.update)#schedule the next update

    def send(self):
        """sends the message

        """
        indexlist = self.li.curselection()
        sendto = []
        for index in indexlist:
            sendto.append(self.namelist[int(index)])#get name from index
        message = self.message.get(1.0, TK.END)#get the message from the box
        for name in sendto:#for each recipient, send the message
            try:
                self.parent.dongle.send_sms(self.parent.database\
                            .get_number(name), str(message))
            except:
                message = "Failed sending message to %s. "\
                                                    % name
                tkMessageBox.showwarning("Message Sending Failed", \
                                          message)
                
        self.window.destroy()


class ViewWindow:
    """The window repsonsible for displaying messages

    """
    def __init__(self, parent, mid):
        width = 485
        height = 220
        #get the message contents
        self.parent = parent
        self.mid = mid
        self.status = self.parent.database.GETMSG_status(mid, client = "View")
        self.date = self.parent.database.GETMSG_date(mid, client = "View")
        self.message = self.parent.database.GETMSG_message(mid, client = "View")
        self.sender = self.parent.database.GETMSG_number(mid, client = "View")
        #build the window contents
        self.window = TK.Toplevel(bg = APP_BG)
        self.window.geometry(centre_me(width = width, height = height))
        self.window.title("Message")
        # From widgets
        TK.Label(self.window, text = "From: ", fg = TITLE_COLOUR, bg = APP_BG)\
                .grid(column = 0,  row = 0)
        TK.Label(self.window, text = self.sender, \
                 fg = ITEM_COLOUR, bg = APP_BG)\
                .grid(column = 1, row = 0, sticky = "w")
        #Status widgets
        TK.Label(self.window, text = " Status: ", fg = TITLE_COLOUR, \
                 bg = APP_BG)\
                .grid(column = 2, row = 0, sticky = "e")
        TK.Label(self.window, text = self.status, fg = ITEM_COLOUR, \
                 bg = APP_BG)\
                .grid(column = 3, row = 0, sticky = "w")
        #Date widgets
        TK.Label(self.window, text = " Date: ", fg = TITLE_COLOUR, \
                 bg = APP_BG)\
                .grid(column = 4, row = 0, sticky = "e")
        TK.Label(self.window, text = self.date,  fg = ITEM_COLOUR, \
                 bg = APP_BG)\
                .grid(column = 5, row = 0, sticky = "w")
        #Message widget
        self.text = TK.Text(self.window, fg = MESSAGE_FG, bg = MESSAGE_BG)
        self.text.insert(TK.END, self.message)
        self.text.config(height = 10,  width = 55)
        self.text.grid(column = 0, row = 1, columnspan = 6, sticky = "e"+"w", \
                       pady = 10)
        self.text.config(state = "disabled")
        #Reply Widget
        TK.Button(self.window, text = "Reply",  command = self.reply, \
        fg = BUTTON_COLOUR, bg = APP_BG).grid(column = 3,  row = 3)
        #Forward Widget
        TK.Button(self.window, text = "Forward", command = self.forward, \
        fg = BUTTON_COLOUR, bg = APP_BG).grid(column = 4, row = 3, padx = 5, )
        #Close Widget
        TK.Button(self.window, text = "Back", \
               command = self.window.destroy, fg = BUTTON_COLOUR, \
               bg = APP_BG)\
              .grid(column = 5, row = 3, padx = 5)
        #Delete Widget
        TK.Button(self.window, text = "Delete", command = self.delete, \
        fg = "red", bg = APP_BG).grid(column = 2, row = 3, padx = 5)
        
    def reply(self):
        """Composes a new message, with the recipient being pre-selected

        """
        recipient = self.parent.database.data["messages"][self.mid]["number"]
        recipient_list = [recipient]
        SendWindow(parent = self.parent.parent, recipients = recipient_list)

    def forward(self):
        """forwards the message onto another person

        """
        message = "FW> "+ self.message
        SendWindow(parent = self.parent.parent, message = message)

    def delete(self):
        """deletes the message from the inbox, and refreshes inbox display

        """
        self.parent.database.delete_message(self.mid)
        self.window.destroy()
        self.parent.hide_row(self.mid)
        self.parent.refresh(sortby= self.parent.sortby, page = self.parent.page)


class Phonebook_Window:
    """the window for adding,displaying,editing and deleting contacts

    """
    def __init__(self, parent, sort = "name"):
        self.sort = sort
        self.parent = parent
        self.database = self.parent.database
        width = 490
        height = 510

        #build the window
        self.window = TK.Toplevel(bg = APP_BG)
        self.window.geometry(centre_me(width = width, height = height))
        self.window.title("Phone Book")
        self.window.bind("<Button-1>", self.cnt_selected)
        self.build_grid(sortby = self.sort, page =1)


    def build_grid(self, page, sortby = "name", reverse = False, filter = None):
        """Builds the phonebook display grid

        """
        anchor = "w"
        self.page = page
        self.sortby = sortby
        self.reverse = reverse
        headwidth = [20, 20, 20]
        # draw the columns
        for col in range(len(self.database.data["contactinfo"])):
            text = self.database.data["contactinfo"][col]
            text = text[:1].upper()+text[1:]
            label = TK.Label(self.window, text = text, width = headwidth[col], \
                      bg = HEADER_BG, anchor = anchor, fg = TITLE_COLOUR)
            label.grid(row = 0,  column = col,  sticky = "w", padx = 10, \
                   pady = 2)
            label.whathit = "HED%s"%text
    #add in the message contents into the rows
        start = (self.page * 20) - 20
        end = start + 20
        row = 1
        self.rowdict = {}
        # get the contacts for that page number
        self.cntlist = self.database.CIDlist(sortby, reverse, filter = filter)

        self.pages = len(self.cntlist) / 20
        if len(self.cntlist) > (self.pages *20):
            self.pages += 1
        slice = self.cntlist[start:end]
        # draw the contacts
        for cid in slice:
            for col in range(len(self.database.data["contactinfo"])):
                colname = self.database.data["contactinfo"][col]          
                text = getattr(self.database, "GETCNT_%s"%colname, \
                               self.database.GET_default)(cid, client = "phone")
                label = TK.Label(self.window, text = text, bg = APP_BG, \
                anchor = anchor, fg = ITEM_COLOUR)
                label.grid(row = row, column = col, sticky = "w", padx = 10)
                label.whathit = "CID%d"% cid# sets the whathit to the row number
                self.rowdict[cid] = row
                
            row += 1

        #adds a few more filler lines if the list is only small
        if row <=MINIMUM_ROWS:
            for num in range((MINIMUM_ROWS-row)+1):
                label = TK.Label(self.window, text = " ", bg = APP_BG)
                label.grid(row = row, column = 0)
                label.whathit = "filler"
                row +=1

        #Display the page number
        TK.Label(self.window, text = " Page %d of %d"%(self.page, self.pages), \
              bg = APP_BG, anchor = anchor,  fg = INFO_COLOUR)\
             .grid(row = row, column = 0, pady = 10, padx = 5, \
                   columnspan = 2, sticky = "w")
        row +=1
        #Display navigation bars.
        label = TK.Label(self.window, text = "<< Previous", bg = APP_BG, \
                  anchor=anchor, fg = NAV_COLOUR)
        label.grid(row = row, column = 0, sticky = "e", padx = 10)
        label.whathit = "NAVNEXT"
        label = TK.Label(self.window, text = "Next >>", bg = APP_BG, \
                  anchor = anchor, fg = NAV_COLOUR)
        label.grid(row = row, column = 1, sticky = "e", padx = 10)
        label.whathit = "NAVPREV"
        row +=1
        #add a tip
        TK.Label(self.window, \
              text = "Tip: click on a contact to edit or delete.", \
              bg = APP_BG, anchor = anchor, fg = INFO_COLOUR)\
             .grid(row = row, column = 0, pady = 10,  padx = 5, \
                   columnspan = 2, sticky = "w")
        row +=1
        #Close the window
        TK.Button(self.window, text = "Back", command = self.window.destroy, \
               fg = "red",  bg = APP_BG).grid(column = 2, row = row, padx = 5)
        #Add contact
        TK.Button(self.window, text = "Add Contact", \
                  command = self.add_contact, fg = "blue", bg = BUTTON_BG)\
                 .grid(column = 0, row = row, padx = 5)

    def refresh(self, page, sortby, sortme = False):
        """Refresh the phonebook display

        """
        # If the column header clicked is the same,  reverse the list
        if sortme == True:
            if sortby.lower() == self.sortby.lower():
                self.reverse = not(self.reverse)
        for item in self.window.grid_slaves():
            item.grid_forget()
        self.build_grid(sortby = sortby, reverse = self.reverse, page = page)

    def cnt_selected(self, event):
        """Event handler for when a message or header is selected

        """
        # set up a dispatcher.
        try : 
            method, params = event.widget.whathit[:3], event.widget.whathit[3:]
        except:
            method, params = "DEF", ""
        getattr(self, "%s_clicked"%method, self.DEF_clicked)(params)

    def CID_clicked(self, params):
        """if a row is clicked,  bring up the contact details screen

        """
        cid = int(params)
        self.parent.master.after(500, self.contact_window(cid=cid, \
                                                          parent = self))

    def contact_window(self, parent, cid):
        """Display the contact details

        """
        ContactWindow(cid = cid, parent = self)

    def HED_clicked(self, params):
        """if a header is clicked,  refresh the srceen.
         Sort by the header clicked
        """
        self.refresh(page = self.page, sortby = params.lower(), sortme = True)

    def NAV_clicked(self, params):
        """What to do if a navigation label is clicked

        """
        if params == "NEXT":
            page = self.page - 1
        elif params == "PREV":
            page = self.page + 1

        if page < 1:
            page = 1# make sure it doesnt go into a minux number
        #make sure the page number isnt greater than the total
        elif page > self.pages:
            page = self.pages
        self.refresh(sortby = self.sortby, sortme = False,  page = page)

    def DEF_clicked(self, params = None):
        """What to do if a non important space is clicked

        """
        pass

    def hide_row(self, cid):
        """Hides a contact entry row, if the contact has been deleted

        """
        for item in self.window.grid_slaves(row = self.rowdict[cid]):
            item.grid_remove()

    def add_contact(self):
        """Builds a contact diplay window, with all the details set to blank
        """
        ContactWindow(self)


class ContactWindow:
    """The window for displaying,editing and deleting a contact

    """
    def __init__(self, parent, cid = None):
        self.cid = cid
        self.parent = parent
        if not self.cid:
            self.name = ""
            self.number = ""
            self.group = ""
        else:
            # where there no info populate with whitespace
            self.name = self.parent.database.GETCNT_name(self.cid, \
                                                         client = "Contact")
            if not self.name:
                self.name = ""
            self.number = self.parent.database.GETCNT_number(self.cid, \
                                                             client = "Contact")
            if not self.number:
                self.number = ""
            self.group = self.parent.database.GETCNT_group(self.cid, \
                                                           client = "Contact")
            if not self.group:
                self.group = ""
        self.parent = parent
        #build the window
        """ labels / entry box for name,  number ,  group
        buttons for save / Exit / Delete - Confirm box for delete
        """
        width = 360
        height = 140
        self.window = TK.Toplevel(bg = APP_BG)
        self.window.geometry(centre_me(width = width, height = height))
        self.window.title("Contact Editing Screen")
        # Name widgets
        TK.Label(self.window, text = "Name: ", fg = TITLE_COLOUR, bg = APP_BG)\
             .grid(column = 0,  row = 0, pady = 10, padx =5, sticky = "w")
        self.nam = TK.Entry(self.window)
        self.nam.config(width = 30)
        self.nam.grid(column = 1,  row = 0, sticky = "w", columnspan = 2)
        self.nam.insert(TK.END, self.name)
        # Telephone widgets
        TK.Label(self.window, text = "Telephone: ", fg = TITLE_COLOUR, \
              bg = APP_BG)\
             .grid(column = 0, row = 1, sticky = "w", pady = 5, padx = 5)
        self.tel = TK.Entry(self.window)
        self.tel.config(width = 30)
        self.tel.grid(column = 1,  row = 1, sticky = "w", columnspan = 2)
        self.tel.insert(TK.END, self.number)
        # Group widgets
        TK.Label(self.window, text = "Group: ", fg = TITLE_COLOUR, bg = APP_BG)\
             .grid(column = 0, row = 2, sticky = "w", pady = 5, padx = 5)
        self.grp = TK.Entry(self.window)
        self.grp.config(width = 30)
        self.grp.grid(column = 1, row = 2, sticky = "w", columnspan = 2)
        self.grp.insert(TK.END, self.group)
        #Save contact button
        if self.cid:
            ttl = "Save Changes"
        else:
            ttl = "Add Contact"
        TK.Button(self.window, text = ttl, command = self.save_contact, \
               fg = "Green", bg = BUTTON_BG)\
              .grid(column = 0, row = 3, padx = 5, pady = 10)
        #Delete contact button
        if self.cid:
            TK.Button(self.window, text = "Delete Contact", \
                   command = self.delete_contact, fg = "blue", bg = BUTTON_BG)\
                  .grid(column = 1, row = 3, padx = 5, pady= 10)
        # Back button
        TK.Button(self.window, text = "Back", command = self.window.destroy, \
               fg = "Red", bg = BUTTON_BG)\
              .grid(column = 2, row = 3, padx = 5, pady = 10, sticky = "e")

    def save_contact(self):
        """Adds or edits a contacts details in the database

        """
        #get the contents of the entry widgets and save them to the data
        name = self.nam.get()
        number = self.tel.get()
        group = self.grp.get()
        self.parent.database.edit_contact(cid = self.cid, name = name, \
                                          number = number, group = group)
        tkMessageBox.showinfo("Saved", "Contact Details Saved")
        self.window.destroy()
        self.parent.refresh(sortby = self.parent.sort, page = self.parent.page)

    def delete_contact(self):
        """Removes a contact from the database

        """
        if tkMessageBox.askokcancel("Delete Contact", "Delete this contact?"):
            self.parent.database.delete_contact(self.cid)
            self.window.destroy()
        self.parent.refresh(sortby = self.parent.sort, page = self.parent.page)


class Dongle:
    """ this class will represent the dongle and will liase with the
        modem class of the humod module
    """
    def __init__(self):
        self.delete_from_sim = False
        try:
            self.m = humod.Modem()
            self.m.enter_text_mode()
            self.online = True
        except:
            self.online = False

    def get_SS(self):
        """finds out the signal stregth of the device and returns it
            as a measure from 0 to 5
        """
        if self.online:
            try:
                strength = self.m.get_rssi()
                if int(strength) == 99:
                    return strength
                strength = int(strength/2.75)# 2.75 appears to be the magic number
                if strength >=5:
                    strength = 5
                return strength
            except:
                return 99
        else: return 99
    def send_sms(self, dest, msg):
        """sends a sms. Takes destination mobile number and message as args.
        returns True if successful,  otherwise returns False
        """
        if self.online:
            self.m.send_text(dest, msg)
            tkMessageBox.showinfo("Message Sent", \
                                " Zoooooooooooooooooooommm!!! ")
        else:
            tkMessageBox.showwarning("Offline Mode",text = \
                    " Unable to send message in offline mode")

    def get_rec_msgs(self):
        """ gets new messages,  returns a nested dict of messages.
        Returned results are deleted from the SIM
        """
        try:
            ml1 = self.m.list_messages()
        except:
            ModemIOError(procedure = "Read", online = self.online)
            self.online = False
            return []
            
        md = {}
        for item in ml1:#build the dictionary for returning
            if len(item) == 5:
                md[item[0]]={"dongle_id":item[0],"status":item[1], \
                             "number":item[2], "date":item[4], \
                             "message":self.m.read_message(item[0])}
                if self.delete_from_sim:
                    self.m.del_message(int(item[0]))
            else: print "ignored - ", item
        return md # return the message dict


class MessageDB:
    """ Stores Phoneboob entries,  messages and provides sorting
        and querying for the classes.
    """
    def __init__(self, dongle):
        self.dongle = dongle
        #try to unpickle the data,  failing that make a new one
        try:
            path = os.path.expanduser("~")
            path = os.path.join(path,".dongler")
        except:
            print "Error : cannot find the user's home directory. exiting"
            sys.exit()
        if not  os.path.isdir(path):
            try:
                os.mkdir(path)
            except:
                print "Error: unable to make the following directory. \
                       Exiting. /n %s " % path
                sys.exit()
        self.datafile = os.path.join(path,"data.data")
        if os.path.isfile(self.datafile):
            try:
                self.data = pickle.load(file(self.datafile))
            except:
                FileIOError(self.datafile,"load")
        else:
            self.data = {}
            self.data["messages"] = {}
            self.data["msgIDcount"] = 1 # count of unique msg IDs
            self.data["CIDcount"] = 1 # count of unique contact IDs
            self.data["contacts"] = {}
            self.data["inboxwindowcolumns"] = {0:"number", 1:"message", \
                    2:"date", 3:"status"}
            self.data["contactinfo"] = {0:"name", 1:"number", 2:"group"}
            self.data["groups"] = ["work", "family", "friends"]
            self.data["new message"] = False

        self.update()
        self.add_names()

    def get_contact_list(self):
        """returns a list of contact names,  in alphabetical order

        """
        li = []
        for cid in self.data["contacts"]:
            name = self.data["contacts"][cid]["name"]
            li.append(name)
            li.sort()
        return li

    def edit_contact(self, name, number, group = None, cid = None):
        """edits a contact in the database and saves the database.
        if the contact doesnt exist,  a new contact/cid is created.
        returns the cid
        """
        if not cid:
            cid = self.data["CIDcount"]
            self.data["CIDcount"] +=1
        self.data["contacts"][cid] = {"name":name, "number":number, \
                                      "group":group}
        self.save_database()
        return cid

    def delete_contact(self, cid):
        """Details a contact from the database

        """
        self.data["contacts"].pop(cid)
        self.save_database()

    # Methods for returning a message contents in a pretty form
    def GETMSG_number(self, mid, client):
        """returns the telephone number from a message
        """
        return self.name_or_number(mid)

    def GETMSG_status(self, mid, client):
        """Returns the status of a message"""
        status = self.data["messages"][mid]["status"]
        if status == "REC UNREAD":
            return "Unread"
        elif status == "REC READ":
            return "Read"
        else:
            return status

    def GETMSG_message(self, mid, client):
        """returns the message portion of the message.
        If the message is too long,  it returns a shortened version
        """
        message = self.data["messages"][mid]["message"]
        if client == "Inbox":
            if len(message) >= 30:
                message = message[:29]+"..."
                return message
            else: return message
        else: return message

    def GETMSG_date(self, mid, client):
        """returns the date field from a message"""
        date = self.data["messages"][mid]["date"]
        date = date[:-6]
        return date

    def GETCNT_name(self, cid, client):
        return self.data["contacts"][cid]["name"]

    def GETCNT_number(self, cid, client):
        return self.data["contacts"][cid]["number"]

    def GETCNT_group(self, cid, client):
        return self.data["contacts"][cid]["group"]

    def GET_default(self, cid, client):
        return "Nothing here"    

    def name_or_number(self, mid):
        """returns a name for a number in a message if exists.
        Otherwise returns number
        """ 
        number = self.data["messages"][mid]["number"]
        if self.data["messages"][mid].has_key("name"):
            return self.data["messages"][mid]["name"]
        else:
            return number

    def get_name(self, number):
        """method for finding a contact name from a telephone number
        note : Returns the first name match,  or None
        """
        for cid in self.data["contacts"]:
            if self.data["contacts"][cid]["number"] == number:
                return self.data["contacts"][cid]["name"]
            else:
                return None

    def has_a_name(self, number):
        """finds if a number has an entry name in the contact list.
            if true,  it returns the name. otherwise it returns False
        """
        for cid in self.data["contacts"]:
            if self.data["contacts"][cid]["number"] == number:
                return self.data["contacts"][cid]["name"]
        return False

    def has_unread_messages(self):
        """ finds out if there are unread messages in the inbox.
            Returns True or False
        """
        for mid in self.data["messages"]:
            if self.data["messages"][mid]["status"] == "REC UNREAD":
                return True
        return False

    def get_number(self, name):
        """finds a contact's telephone number using their name
        if the name cannot be found,  Returns None
        """
        for cid in self.data["contacts"]:
            if self.data["contacts"][cid]["name"] == name:
                return self.data["contacts"][cid]["number"]
        else: return None

    def get_msg(self, msg_ID):
        """tries to find a message from its id number.Returns the message dict
            If it fails,  returns None
        """
        if msg_ID in self.data["messages"]: return self.data["messages"][msg_ID]
        else: return None

    def set_msg_status(self, msg_ID, status):
        """sets the status of a message to status

        """
        self.data["messages"][msg_ID]["status"] = status
        self.save_database()

    def is_a_duplicate(self,newmsg):
        for msg in self.data["messages"]:
            if self.data["messages"][msg]["dongle_id"] == \
               newmsg["dongle_id"]:
                if self.data["messages"][msg]["date"] == \
                   newmsg["date"]:
                    return True
        return False

    def dongle_id_list(self):
        dlist = []
        for msg in self.data["messages"]:
              dongle_id = int(self.data["messages"][msg]["dongle_id"])
              dlist.append(dongle_id)
        return dlist

    def add_names(self):
        """adds names to messages in the database,  from the phonebook"""
        for msgid in self.data["messages"]:
            name = self.has_a_name((self.data["messages"][msgid]["number"]))
            if name:
                self.data["messages"][msgid]["name"] = name
                            
    def update(self):
        """ gets new messages from the dongle and updates the database"""
        dict = self.dongle.get_rec_msgs()
        dlist = self.dongle_id_list()
        for key in dict:
            if dict[key]["dongle_id"] in dlist:
                if self.is_a_duplicate(dict[key]):
                    continue
            else:
                self.data["new message"] = True
                msgID = self.data["msgIDcount"]
                self.data["messages"][msgID] = dict[key]
                self.data["msgIDcount"] +=1 # increase the count
        self.add_names()
        self.save_database()

    def invert_dict(self):
        """inverts the contacts dictionary"""
        inv = {}
        for item in self.data["contacts"]:
            key = self.data["contacts"][item]["number"]
            inv[key] = item
        return inv

    def save_database(self):
        """saves the database contents to the database file"""
        try:
            datafile = file(self.datafile, 'w')
            pickle.dump(self.data, datafile)
        except:
            FileIOError(file,"save")

    def messageIDlist(self, sortby = "date", reverse = False,  filter = None):
        """returns a list of message IDs from the database.
        Sorted and filtered as specified"""

        if filter:# first do any filtering
            if not type(filter) == type(tuple()):# Bit or error checking
                print "Supplied filter argument is not a tuple.returning all"
                data = self.data["messages"]
            else:
                data = {}
                for key in self.data["messages"]:
                    for status in filter:
                        if self.data["messages"][key]["status"] == status:
                            data[key] = self.data["messages"][key]
        else: data = self.data["messages"]
        # Then do the sorting
        li = []
        for MSGID in data:
            li.append((data[MSGID][sortby], MSGID))
        
        li.sort()# sort the resultant list
        if reverse:
            li.reverse()
        return [index[1] for index in li]



    def CIDlist(self, sortby = "name", reverse = False,  filter = None):
        """returns a list of Contact IDs from the database.
        Sorted and Filtered as specified
        """

        if filter:# a Vaild filter arguement would be a group name
            if not type(filter) == type(tuple()):# Bit or error checking
                print "Supplied filter argument is not a tuple.Returning all"
                data = self.data["messages"]
            else:
                data = {}
                for cid in self.data["contacts"]:
                    for group in filter:
                        if self.data["contacts"][cid]["group"] == group:
                            data[cid] = self.data["messages"][cid]
        else: data = self.data["contacts"]
        # Then do the sorting
        li = []
        for cid in data:
            li.append((data[cid][sortby], cid))
        
        li.sort()# sort the resultant list
        if reverse:
            li.reverse()
        return [index[1] for index in li]

    def delete_message(self, mid):
        """changes the message status flag to DELETED"""
        self.data["messages"][mid]["status"] = "DELETED"


class FileIOError():
    def __init__(self, file, procedure):
        text = "FileIOException: Unable to %s the  %s" % (procedure, file)
        tkMessageBox.showwarning("Error",text) 
        sys.exit()

class ModemIOError():
    def __init__(self,procedure,online):
        pass

class ImportError():
    def __init__(self):
        print "unable to import a required modules."
        sys.exit()


def centre_me(height, width):
    """returns geometry info for the window"""
    coordx = (SWIDTH/2) - (width/2)
    coordy = (SHEIGHT/2) - (height/2)
    return ("%dx%d+%d+%d")%(width, height, coordx, coordy)


def run():
        ROOT = TK.Tk()
        ROOT.configure(bg = APP_BG)
        DONGLE = Dongle()
        DATABASE = MessageDB(DONGLE)
        global SWIDTH
        global SHEIGHT
        # get the size of the screen
        SWIDTH = ROOT.winfo_screenwidth()
        SHEIGHT = ROOT.winfo_screenheight()
        ROOT.geometry(centre_me(width = WW, height = WH))
        ROOT.title("Dongler")
        #build the app,  giving the root window as master and the dongle.
        APP = MainWindow(ROOT, DONGLE, DATABASE)
        ROOT.mainloop()

if __name__ == "__main__":
    run()
